""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import re
import pika
import time

from django import db
from django.db.utils import InterfaceError, OperationalError
from utils.exchange_validator_utility import update_router_state
from utils.logger import get_logger
from dispatcher.environment import expand
from utils import globals
from django.conf import settings
from retry import retry
from utils.utility import get_requests_auth, api_call
from utils.rmq_utility import prepare_inter_cyops_config, prepare_intra_cyops_config, create_exchange, \
    get_rmq_connection, prepare_health_check_inter_config, health_check as check_agent_health, update_tenant_status,\
    health_check as check_sme_health

from utils.callbacks import get_callback
from utils.constants import *

from utils.cluster_utils import is_active_primary, get_primary_hostname_for_embedded_sme, get_self_hostname

from dispatcher.callback.data.rabbitmq import Client
from utils.rmq_utility import create_vhost_and_user_if_not_exists

from utils.utility import get_router_details_from_iri

logger = get_logger(__name__)


class Consumer:
    def __init__(self, config):
        self.config = config

    def __enter__(self):
        self.callback_map = dict()
        self.connection = self._create_connection()
        self.channel = self.connection.channel()
        if self.config.get('prefetch_count'):
            self.channel.basic_qos(prefetch_count=self.config.get('prefetch_count'))
        # self._create_exchange(self.channel)
        return self

    def __exit__(self, *args):
        self.connection.close()

    def prepare_consumers(self, queue_config, message_received_callback):
        logger.info('Preparing Consumer for basic consume')
        self.callback_map.update({queue_config['routingKey']: message_received_callback})

        self._create_queue(self.channel, queue_config)

        self.channel.queue_bind(queue=queue_config['queueName'],
                                exchange=queue_config['exchangeName'],
                                routing_key=queue_config['routingKey'])

        self.channel.basic_consume(on_message_callback=self._consume_message, queue=queue_config['queueName'])

    def start_consuming(self):
        logger.info('Starting consumer')
        self.channel.start_consuming()

    def stop_consuming(self):
        logger.info('Stopping requested consumer')
        try:
            self.channel.stop_consuming()
            globals.consumer_pool.pop(self.config.get('id'), None)
            globals.health_check_consumer_pool.pop(self.config.get('id'), None)
            logger.info('Stopped requested consumer successfully')
        except Exception as e:
            logger.error('Error while stopping consumer %s', str(e))

    def _create_exchange(self, channel):
        logger.info('Creating exchange: %s', self.config['exchangeOptions'])
        exchange_options = self.config['exchangeOptions']
        channel.exchange_declare(exchange=self.config['exchangeName'],
                                 exchange_type=self.config['exchangeType'],
                                 passive=exchange_options['passive'],
                                 durable=exchange_options['durable'],
                                 auto_delete=exchange_options['autoDelete'],
                                 internal=exchange_options['internal'])

    def _create_queue(self, channel, queue_config):
        logger.info('Creating queue: %s', queue_config['queueName'])
        queue_options = queue_config['queueOptions']
        channel.queue_declare(queue=queue_config['queueName'],
                              passive=queue_options['passive'],
                              durable=queue_options['durable'],
                              exclusive=queue_options['exclusive'],
                              auto_delete=queue_options['autoDelete'])

    def _create_connection(self):
        logger.info('Creating connection')
        return get_rmq_connection(**{'config': self.config})

    def _consume_message(self, channel, method, properties, body):
        logger.info('Consuming message and calling desired callback')
        callback = self.callback_map.get(method.routing_key)
        if not callback:
            for key, value in self.callback_map.items():
                if re.match(key, method.routing_key):
                    callback = value
                    break
        _retry = 2  # Max attempt in any scenario - Two times
        while _retry > 0:
            try:
                callback(**{'message': body, 'method': method, 'properties': properties})
            except (InterfaceError, OperationalError):
                # When data base unexpectedly closes connection due to it's server restart or timeout django does not
                # handle it gracefully and django assumes that connection is never closed and when we access it raises
                # interface error exception as connection already closed.
                logger.warn('Django database connection closed detected retrying.')
                db.connection.close()
                _retry -= 1
                continue
            except Exception as e:
                logger.exception('Exception Trace for callback function %s: ', str(e))
            _retry = 0

        channel.basic_ack(delivery_tag=method.delivery_tag)


@retry(delay=settings.ROUTER_RETRY_DELAY, logger=logger)
def start_health_check_consumer(router, *args, **kwargs):
    try:
        router_uuid = router.get('@id').split('/')[-1]
        router = settings.ROUTERS.get(router.get('@id'), {})

        if not router and globals.health_check_consumer_pool.get(router_uuid) and \
                globals.health_check_consumer_pool.get(router_uuid).connection.is_open:
            return

        # need not setup health check for 'Master' router at tenant. it doesn't have admin creds
        if router.get('@id') == master_router_at_tenant:
            return

        # Default config status is "Not Configured"
        config_status = settings.TENANT_CONFIG_STATUS.get('0').get('iri')

        # Router health is a dict which stores nodeid against the configStatus
        self_hostname = get_self_hostname()
        if router.get("health") and router.get('health').get(self_hostname):
            config_status = router.get('health').get(self_hostname).get('configStatus')

        # Not creating exchange as exchange is always fixed
        config = prepare_health_check_inter_config(router)

        if not check_sme_health(config):
            config_message = 'Invalid credentials or certificate.'
            if not config_status or config_status != settings.TENANT_CONFIG_STATUS.get('-1').get('iri'):
                update_router_state(router.get('@id'), settings.TENANT_CONFIG_STATUS.get('-1').get('iri'), config_message)
            raise Exception()
        else:
            config_message = 'Configuration successful'
            if not config_status or config_status != settings.TENANT_CONFIG_STATUS.get('2').get('iri'):
                update_router_state(router.get('@id'), settings.TENANT_CONFIG_STATUS.get('2').get('iri'), config_message)

        with Consumer(config) as consumer:
            logger.info('Starting consumer for APP %s', config.get('id'))
            globals.health_check_consumer_pool[config.get('id')] = consumer
            for each_event in config.get('event_listener'):
                message_received_callback = get_callback(each_event)
                queue_config = {}
                each_event.update({'queueName': each_event['queueName'].replace('{{self_id}}', settings.SELF_ID)})
                queue_config.update(each_event)
                try:
                    consumer.prepare_consumers(queue_config, message_received_callback)
                except pika.exceptions.AuthenticationError:
                    globals.health_check_consumer_pool.pop(config.get('id'), None)
                    # Adding as for dedicated tenant this mq entry will be there but password wont be there for admin
                    # Hence will get auth error.
                    logger.warn('\nAuthentication error, for message exchange\n')
                    return
                except pika.exceptions.ChannelClosed:
                    # Returning in exception as exchange exception will be same for all the events
                    globals.health_check_consumer_pool.pop(config.get('id'), None)
                    logger.warn(
                        '\nExchange `amq.rabbitmq.event` does not exist, enable `rabbitmq_event_exchange`'
                        'in your secure message exchange  using `rabbitmq-plugins enable rabbitmq_event_exchange`\n')
                    logger.warn('\nDo ensure your rabbitmq-server version is 3.8.* or higher\n')
                    return
                except Exception as e:
                    globals.health_check_consumer_pool.pop(config.get('id'), None)
                    logger.warn('Exception occurred while configuring monitor router %s', str(e))
                    return

            consumer.start_consuming()
    except Exception as err:
        logger.exception(err)
        raise err


@retry(delay=settings.ROUTER_RETRY_DELAY, logger=logger)
def start_consumer(agent_id, *args, **kwargs):
    try:
        if globals.consumer_pool.get(agent_id) and globals.consumer_pool.get(agent_id).connection.is_open:
            return
        agent = settings.TENANTS.get(agent_id)
        if not agent or not agent.get('active'):
            logger.debug(f"Agent {agent_id} not active. Returning")
            return

        if isinstance(agent.get('configstatus'), dict):
            configstatus = agent.get('configstatus').get('@id')
        else:
            configstatus = agent.get('configstatus')

        is_agent_configured = agent.get('role') == 'self' or configstatus in [
            settings.TENANT_CONFIG_STATUS.get('2').get('iri'),
            settings.TENANT_CONFIG_STATUS.get('3').get('iri'),
            settings.TENANT_CONFIG_STATUS.get('4').get('iri')
        ]

        if settings.IS_MASTER and not is_agent_configured and agent.get('role') != 'self':
            logger.warn('Agent %s is not configured. configuring agent', agent_id)

        if agent.get('role') == 'self':
            config = prepare_intra_cyops_config(agent_id)
        else:
            if not is_active_primary():
                logger.debug(f"Not starting consumer for {agent_id} since this is not primary node")
                return
            else:
                logger.debug(f"Starting consumer for {agent_id}")
            config = prepare_inter_cyops_config(agent)

        # Check if vhost exists for the given agent.
        # If not create vhost and user for this agent.
        # This case could happen when the vhost gets deleted
        if agent.get('router'):
            if type(agent.get('router')) != dict:
                admin_router = get_router_details_from_iri(agent.get('router'))
            else:
                admin_router = agent.get('router')
            # Here need to check if the below code is run on the tenant and the router is master's.
            # In this case we need to skip the below create_vhost_and_user is not required to be run
            if not admin_router.get('@id') == master_router_at_tenant:

                router_type = admin_router.get('type', {}).get('@id') if isinstance(admin_router.get('type'), dict) else admin_router.get('type')

                host = admin_router.get('address')
                if router_type == fsr_embedded:
                    host = get_primary_hostname_for_embedded_sme()

                sni = None
                if host != get_self_hostname():
                    sni = admin_router.get('sni') if admin_router.get('sni') else admin_router.get('address')


                rabbitmq_client = Client(host + ':' + str(admin_router.get('apiPort')) + '/api',
                                         admin_router.get('username'), admin_router.get('password'),
                                         ca_cert=admin_router.get('certificate').encode(), address=host,
                                         port=admin_router.get('apiPort'), sni=sni)

                agent_router = {}
                agent_router['vhost'] = agent.get('vhost')
                agent_router['username'] = agent.get('username')
                agent_router['password'] = agent.get('password')
                agent_router['client_cert'] = agent.get('clientCert')
                agent_router['auth_type'] = agent.get('authType', {}).get('@id') if isinstance(agent.get('authType', {}),
                                                                                               dict) else agent.get('authType')
                create_vhost_and_user_if_not_exists(rabbitmq_client, agent_router)

        # check agent health before proceeding
        if not check_agent_health(config):
            if configstatus != settings.TENANT_CONFIG_STATUS.get('4').get('iri'):
                update_tenant_status(config.get('tenant_iri'), settings.TENANT_CONFIG_STATUS.get('4').get('iri'),
                                     failed_on_message_exchange)
            raise Exception()

        create_exchange(config, settings.SELF_ID)
        # keeping consumer's prefetch count to 4 so that consumers buffer is not filled with many messages
        config['prefetch_count'] = 4
        with Consumer(config) as consumer:
            logger.info('Starting consumer for APP %s', config.get('id'))
            globals.consumer_pool[config.get('id')] = consumer
            for each_queue in config.get('rabbit', {}).get('queues'):
                each_queue = {key: expand({'self_id': settings.SELF_ID}, value) for key, value in each_queue.items()}
                message_received_callback = get_callback(each_queue)
                queue_config = {}
                queue_config.update(each_queue)
                consumer.prepare_consumers(queue_config, message_received_callback)
            consumer.start_consuming()
    except Exception as err:
        logger.exception(err)
        raise err


if __name__ == '__main__':
    pass
